-- Author: Erenst Jamro - All Rights Riserved
-- Utility to generate clocks within FPGAs

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

library UNISIM;
use unisim.all; -- for global set reset signal

entity util_clkgen is
  generic (c_clk_in_period_ns: real:= 0.0; -- clk_input period
           c_external_feedback: integer:= 1; -- -1 No feedback at all, 0- internal, 1- external
           c_clk_feedback1X: integer:= 1; -- clk output frequency feedback= 0 no feedback, =1 is the same as clk_input, =2 clk_out frequency is twice the clk_input frequency
           c_clkdv_divide: real:= 1.0; -- 
           c_fxmultiply: integer:= 0; -- clock 0 multiply factor
           c_fxdivide: integer:= 5); -- clock 0 multiply factor
  port ( clk_input: in std_logic;  -- input clock
         clk_external_o, clk_external_t, clk_internal, clk_fx, clk_fx_external: out std_logic; -- the first clock output (external or internal signal)
         clk_dv, clk_dv_external: out std_logic; -- clock = clk / c_clkdv_divide - used only for c_clk_divide>1.0
         clk_external_i : in std_logic); -- the first clock input (from the external pin
end util_clkgen;

architecture arch of util_clkgen is     
component IBUFG  -- special input buffer for clock distribution
   port(
      O                              :	out   STD_ULOGIC;
      I                              :	in    STD_ULOGIC);
   end component;
component BUFG
   port(
      O                              :	out   STD_ULOGIC;
      I                              :	in    STD_ULOGIC);
   end component;

component DCM
    generic (DFS_FREQUENCY_MODE : string := "LOW";
             DLL_FREQUENCY_MODE : string := "LOW";
             DUTY_CYCLE_CORRECTION : boolean := TRUE;
             CLKIN_DIVIDE_BY_2 : boolean := FALSE;
             CLK_FEEDBACK : string := "1X";
             CLKOUT_PHASE_SHIFT : string := "NONE";
             FACTORY_JF : bit_vector := X"C080";
             STARTUP_WAIT : boolean := TRUE;
	           DSS_MODE	  : string := "NONE";
             PHASE_SHIFT  : integer := 0 ;
             CLKFX_MULTIPLY : integer  := 4 ;
	           CLKFX_DIVIDE : integer  := 1;
             CLKDV_DIVIDE : real := 2.0;
             CLKIN_PERIOD : real := 0.0;                 
             DESKEW_ADJUST : string := "SYSTEM_SYNCHRONOUS"
             ); 
    port (
          CLK0    : out std_ulogic := '0';
          CLK180  : out std_ulogic := '0';
          CLK270  : out std_ulogic := '0';
          CLK2X   : out std_ulogic := '0';
          CLK2X180 : out std_ulogic := '0';
          CLK90   : out std_ulogic := '0';          
          CLKDV   : out std_ulogic := '0';
          CLKFX   : out std_ulogic := '0';
          CLKFX180 : out std_ulogic := '0';
          LOCKED  : out std_ulogic := '0';
          PSDONE  : out std_ulogic := '0';
          STATUS  : out std_logic_vector(7 downto 0) := "00000000";
          CLKFB   : in  std_ulogic := '0';          
          CLKIN   : in  std_ulogic := '0';
          DSSEN : in  std_ulogic := '0';
          PSCLK : in  std_ulogic := '0';
          PSEN : in  std_ulogic := '0';          
          PSINCDEC : in  std_ulogic := '0';
          RST     : in  std_ulogic := '0');
end component;
component SRL16
  generic (
       INIT : bit_vector := X"0000"
  );
  port (
        Q   : out STD_ULOGIC;
        A0  : in STD_ULOGIC;
        A1  : in STD_ULOGIC;
        A2  : in STD_ULOGIC;
        A3  : in STD_ULOGIC;
        CLK : in STD_ULOGIC;        
        D   : in STD_ULOGIC
       ); 
   end component;
Function FindCLKFeedBack(clk_feedback: integer) return string is begin
  if clk_feedback<=0 then return "None";
  elsif clk_feedback=1 then return "1X";
  else return "2X";
  end if;
end function;
Function FindCLKDV_divide(clkdv_divide: real) return real is begin -- clkdv_divide can not be less than 1.5
  if clkdv_divide <= 1.5 then return 2.0;
  else return clkdv_divide;
  end if;
end function;
 constant clk_feedback: string:= FindCLKFeedBack(c_clk_feedback1X);
 constant clkdv_divide: real:= FindCLKDV_divide(c_clkdv_divide);
    function MAX(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0>=arg1 then ret:= arg0;
    else ret:= arg1;
    end if;
    return ret;
  end MAX;
  constant fxmultiply: integer:= MAX(2, c_fxmultiply); 
  signal clk_fb, clk1x, clk2x, clk12x, clk_fx1, clk_fx2: std_logic; -- feed-back signal
  signal clk_dv1, clk_dv2: std_logic; -- clock divided by a constant value
  signal rst: std_logic;
begin

  g_nobuf: if c_external_feedback<0 generate -- no global buffer
    clk_fb<= clk12x;
    clk_external_o<= clk_fb;
    rst<= '0';
  end generate;

  g0int: if c_external_feedback=0 generate
    gl: BUFG
      port map(O=> clk_fb, I=> clk12x);
    clk_external_o<= clk_fb;
    rst<= '0';
  end generate;
  g0ext: if c_external_feedback>0 generate
    --ib: IBUFG 
    --  port map(O=> clk_fb,  I=> clk_external_i);
    clk_fb<= clk_external_i;
    clk_external_o<= clk12x;
    reset: SRL16
      generic map(INIT=> X"FFFF")
      port map(Q=> rst, A0=> '1', A1=> '1', A2=> '1', A3=> '1',
        CLK=> clk_input, D=> '0');
  end generate;
  g1dv: if c_clkdv_divide > 1.0 generate -- use clk divide logic
    gl: BUFG
      port map(O=> clk_dv2, I=> clk_dv1);
    clk_dv<= clk_dv2;
    clk_dv_external<= clk_dv2;
  end generate;
  
  clk_internal<= clk_fb;
  
  f0: if c_clk_feedback1X<=0 generate
    clk12X<= '0';
  end generate; 
  
  f1: if c_clk_feedback1X=1 generate
    clk12X<= clk1X;
  end generate;
  
  f2: if c_clk_feedback1X>1 generate
    clk12X<= clk2X;
  end generate;
  
  fx1: if c_fxmultiply>=2 generate
    gl: BUFG
      port map(O=> clk_fx2, I=> clk_fx1);
    clk_fx<= clk_fx2;
    clk_fx_external<= clk_fx2;
  end generate;
  
  c0: DCM
    generic map (DFS_FREQUENCY_MODE=> "LOW",
             DLL_FREQUENCY_MODE=> "LOW",
             DUTY_CYCLE_CORRECTION=> TRUE,
             CLKIN_DIVIDE_BY_2=> FALSE,
             CLK_FEEDBACK=> clk_feedback,
             CLKOUT_PHASE_SHIFT=> "NONE",
             FACTORY_JF=> X"C080",
             STARTUP_WAIT=> TRUE,
	           DSS_MODE=> "NONE",
             PHASE_SHIFT=> 0,
             CLKFX_MULTIPLY=> fxmultiply,
	           CLKFX_DIVIDE=> c_fxdivide,
             CLKDV_DIVIDE=> clkdv_divide,
             CLKIN_PERIOD=> c_clk_in_period_ns,                 
             DESKEW_ADJUST=> "SYSTEM_SYNCHRONOUS" )
     port map(CLK0=> clk1X,
          CLK180=> open, CLK270=> open, CLK2X=> clk2x, CLK2X180=> open, CLK90=> open, CLKDV=> clk_dv1,
          CLKFX=> clk_fx1, CLKFX180=> open, LOCKED=> open, PSDONE=> open, STATUS=> open, 
          CLKFB=> clk_fb, CLKIN=> clk_input, RST=> rst);
  
    clk_external_t<= '0';      
end arch;  